import logging
import logging.handlers
import os
from urllib.parse import urlparse


def create_formatter(logging_metadata):
    """Create a log formatter from the given logging metadata that delimits the context fields with space-padded
    pipes and encapsulates the whole context in square brackets before adding the message at the end. e.g. if the
    logging metadata was `("%(asctime)s", "%(levelname)s", "%(name)s")`, the formatter would format log messages as e.g.
    `[2021-06-29 11:58:10,985 | INFO | octue.runner] This is a log message.`

    :param iter(str) logging_metadata: an iterable of context fields to use as context for every log message that the formatter is applied to
    :return logging.Formatter:
    """
    return logging.Formatter("[" + " | ".join(logging_metadata) + "]" + " %(message)s")


# Logging format for analysis runs. All handlers should use this logging format, to make logs consistently parseable
LOGGING_METADATA_WITH_TIMESTAMP = ("%(asctime)s", "%(levelname)s", "%(name)s")
FORMATTER_WITH_TIMESTAMP = create_formatter(LOGGING_METADATA_WITH_TIMESTAMP)
FORMATTER_WITHOUT_TIMESTAMP = create_formatter(LOGGING_METADATA_WITH_TIMESTAMP[1:])


def apply_log_handler(logger_name=None, handler=None, log_level=logging.INFO, formatter=None):
    """Apply a log handler with the given formatter to the logger with the given name.

    :param str|None logger_name: if this is `None`, the root logger is used
    :param logging.Handler handler: The handler to use. If None, default console handler will be attached.
    :param int log_level: ignore log messages below this level
    :param logging.Formatter|None formatter: if this is `None`, the default `FORMATTER_WITH_TIMESTAMP` is used
    :return logging.Logger:
    """
    handler = handler or logging.StreamHandler()
    handler.setFormatter(formatter or FORMATTER_WITH_TIMESTAMP)
    handler.setLevel(log_level)
    logger = logging.getLogger(name=logger_name)
    logger.addHandler(handler)
    logger.setLevel(log_level)

    for handler in logger.handlers:
        if type(handler).__name__ == "SocketHandler":
            # Log locally that a remote logger will be used.
            local_logger = logging.getLogger(__name__)
            temporary_handler = logging.StreamHandler()
            temporary_handler.setFormatter(formatter or FORMATTER_WITH_TIMESTAMP)
            temporary_handler.setLevel(log_level)
            local_logger.addHandler(temporary_handler)
            local_logger.setLevel(log_level)
            local_logger.info("Logs streaming to %s:%s", handler.host, str(handler.port))
            local_logger.removeHandler(temporary_handler)
            break

    return logger


def get_remote_handler(logger_uri, formatter=None):
    """Get a log handler for streaming logs to a remote URI accessed via HTTP or HTTPS. The given formatter is applied.

    :param str logger_uri: the URI to stream the logs to
    :param logging.Formatter|None formatter: if this is `None`, the `FORMATTER_WITH_TIMESTAMP` formatter is used
    :return logging.Handler:
    """
    parsed_uri = urlparse(logger_uri)

    if parsed_uri.scheme not in {"ws", "wss"}:
        raise ValueError(
            f"Only WS and WSS protocols currently supported for remote logger URI. Received {logger_uri!r}."
        )

    handler = logging.handlers.SocketHandler(host=parsed_uri.hostname, port=parsed_uri.port)
    handler.setFormatter(formatter or FORMATTER_WITH_TIMESTAMP)
    return handler


def get_formatter():
    """Get the correct formatter for the environment. If the environment is Google Cloud Run, use a log handler with a
    formatter that doesn't include the timestamp in the log message context to avoid the date appearing twice in the
    Google Cloud Run logs (Google adds its own timestamp to log messages).

    :return logging.Formatter:
    """
    if os.environ.get("COMPUTE_PROVIDER", "UNKNOWN") == "GOOGLE_CLOUD_RUN":
        return FORMATTER_WITHOUT_TIMESTAMP
    else:
        return FORMATTER_WITH_TIMESTAMP



import uuid
from unittest.mock import patch

from octue.resources.child import Child
from octue.resources.service_backends import GCPPubSubBackend
from tests.base import BaseTestCase
from tests.cloud.pub_sub.mocks import MockAnalysis, MockService, MockSubscriber, MockSubscription, MockTopic


class TestChild(BaseTestCase):
    def test_child_can_be_asked_multiple_questions(self):
        """Test that a child can be asked multiple questions."""
        backend = GCPPubSubBackend(project_name="blah")

        def run_function(analysis_id, input_values, input_manifest, analysis_log_handler):
            return MockAnalysis(output_values=input_values)

        responding_service = MockService(backend=backend, service_id=str(uuid.uuid4()), run_function=run_function)

        with patch("octue.cloud.pub_sub.service.Topic", new=MockTopic):
            with patch("octue.cloud.pub_sub.service.Subscription", new=MockSubscription):
                with patch("octue.resources.child.BACKEND_TO_SERVICE_MAPPING", {"GCPPubSubBackend": MockService}):
                    with patch("google.cloud.pubsub_v1.SubscriberClient", new=MockSubscriber):
                        responding_service.serve()

                        child = Child(
                            name="wind_speed",
                            id=responding_service.id,
                            backend={"name": "GCPPubSubBackend", "project_name": "blah"},
                        )

                        # Make sure the child's underlying mock service knows how to access the mock responding service.
                        child._service.children[responding_service.id] = responding_service
                        self.assertEqual(child.ask([1, 2, 3, 4])["output_values"], [1, 2, 3, 4])
                        self.assertEqual(child.ask([5, 6, 7, 8])["output_values"], [5, 6, 7, 8])



def create_exceptions_mapping(*sources):
    """Create a single mapping of exception names to their classes given any number of dictionaries mapping variable
    names to variables e.g. `locals()`, `globals()` or a module. Non-exception variables are filtered out. This function
    can be used to combine several modules of exceptions into one mapping.

    :param sources: any number of `dict`s of global or local variables mapping object names to objects
    :return dict:
    """
    candidates = {key: value for source in sources for key, value in source.items()}

    exceptions_mapping = {}

    for name, object in candidates.items():
        try:
            if issubclass(object, BaseException):
                exceptions_mapping[name] = object

        except TypeError:
            continue

    return exceptions_mapping



from unittest.mock import patch
import google.api_core.exceptions
from google.pubsub_v1 import SubscriberClient

from octue.cloud.pub_sub.service import Service
from octue.cloud.pub_sub.subscription import SEVEN_DAYS, THIRTY_ONE_DAYS, Subscription
from octue.cloud.pub_sub.topic import Topic
from octue.resources.service_backends import GCPPubSubBackend
from tests import TEST_PROJECT_NAME
from tests.base import BaseTestCase
from tests.cloud.pub_sub.mocks import MockSubscriber


class TestSubscription(BaseTestCase):

    service = Service(backend=GCPPubSubBackend(project_name="my-project"))
    topic = Topic(name="world", namespace="hello", service=service)
    subscription = Subscription(
        name="world", topic=topic, namespace="hello", project_name=TEST_PROJECT_NAME, subscriber=MockSubscriber()
    )

    def test_namespace_only_in_name_once(self):
        """Test that the subscription's namespace only appears in its name once, even if it is repeated."""
        self.assertEqual(self.subscription.name, "hello.world")

        subscription_with_repeated_namespace = Subscription(
            name="hello.world",
            topic=self.topic,
            namespace="hello",
            project_name=TEST_PROJECT_NAME,
            subscriber=MockSubscriber(),
        )

        self.assertEqual(subscription_with_repeated_namespace.name, "hello.world")

    def test_repr(self):
        """Test that Subscriptions are represented correctly."""
        self.assertEqual(repr(self.subscription), "<Subscription(hello.world)>")

    def test_error_raised_when_creating_without_allowing_existing_when_subscription_already_exists(self):
        """Test that an error is raised when trying to create a subscription that already exists and `allow_existing` is
        `False`.
        """
        with patch(
            "tests.cloud.pub_sub.mocks.MockSubscriber.create_subscription",
            side_effect=google.api_core.exceptions.AlreadyExists(""),
        ):
            with self.assertRaises(google.api_core.exceptions.AlreadyExists):
                self.subscription.create(allow_existing=False)

    def test_create_with_allow_existing_when_already_exists(self):
        """Test that trying to create a subscription that already exists when `allow_existing` is `True` results in no
        error.
        """
        with patch(
            "tests.cloud.pub_sub.mocks.MockSubscriber.create_subscription",
            side_effect=google.api_core.exceptions.AlreadyExists(""),
        ):
            self.subscription.create(allow_existing=True)

    def test_create(self):
        """Test that creating a subscription works properly."""
        service = Service(backend=GCPPubSubBackend(project_name=TEST_PROJECT_NAME))
        topic = Topic(name="my-topic", namespace="tests", service=service)
        subscription = Subscription(
            name="world",
            topic=topic,
            namespace="hello",
            project_name=TEST_PROJECT_NAME,
            subscriber=SubscriberClient(credentials=service._credentials),
        )

        try:
            topic.create(allow_existing=True)
            response = subscription.create(allow_existing=True)

        finally:
            try:
                subscription.delete()
            finally:
                topic.delete()

        self.assertEqual(response._pb.ack_deadline_seconds, 60)
        self.assertEqual(response._pb.expiration_policy.ttl.seconds, THIRTY_ONE_DAYS)
        self.assertEqual(response._pb.message_retention_duration.seconds, SEVEN_DAYS)



import json
import os
import shutil
import subprocess
import sys
import tempfile
import time
import uuid

from octue import Runner
from octue.utils.processes import ProcessesContextManager
from ..base import BaseTestCase


class TemplateAppsTestCase(BaseTestCase):
    """ Test case that runs analyses using apps in the templates, to ensure all the examples work. """

    def setUp(self):
        super().setUp()
        self.start_path = os.getcwd()

        # Initialise so these variables are assigned on the instance
        self.template_data_path = None
        self.template_twine = None
        self.template_path = None
        self.app_test_path = None
        self.teardown_templates = []

        def set_template(template):
            """Sets up the working directory and data paths to run one of the provided templates"""
            self.template_path = os.path.join(self.templates_path, template)
            self.template_twine = os.path.join(self.templates_path, template, "twine.json")

            # Duplicate the template's data/ directory to a test-specific replica
            self.app_test_path = os.path.join(self.data_path, str(uuid.uuid4()))
            shutil.copytree(self.template_path, self.app_test_path)

            # Add this template to the list to remove in teardown
            self.teardown_templates.append(self.app_test_path)
            sys.path.insert(0, self.app_test_path)

            # Run from within the app folder context
            os.chdir(self.app_test_path)

        self.set_template = set_template

    def tearDown(self):
        """Remove data directories manufactured"""
        super().tearDown()
        os.chdir(self.start_path)
        for path in self.teardown_templates:
            sys.path.remove(path)
            shutil.rmtree(path)

    def test_fractal_configuration(self):
        """ Ensures fractal app can be configured with its default configuration. """
        self.set_template("template-python-fractal")
        runner = Runner(
            app_src=self.template_path,
            twine=self.template_twine,
            configuration_values=os.path.join("data", "configuration", "configuration_values.json"),
            output_manifest_path=os.path.join("data", "output", "manifest.json"),
        )

        analysis = runner.run()
        analysis.finalise(output_dir=os.path.join("data", "output"))

    def test_using_manifests(self):
        """ Ensures using-manifests app works correctly. """
        self.set_template("template-using-manifests")
        runner = Runner(
            app_src=self.template_path,
            twine=self.template_twine,
            configuration_values=os.path.join("data", "configuration", "values.json"),
            output_manifest_path=os.path.join("data", "output", "manifest.json"),
        )

        analysis = runner.run(input_manifest=os.path.join("data", "input", "manifest.json"))
        analysis.finalise(output_dir=os.path.join("data", "output"))
        self.assertTrue(os.path.isfile(os.path.join("data", "output", "cleaned_met_mast_data", "cleaned.csv")))

    def test_child_services_template(self):
        """Ensure the child services template works correctly (i.e. that children can be accessed by a parent and data
        collected from them). This template has a parent app and two children - an elevation app and wind speed app. The
        parent sends coordinates to both children, receiving the elevation and wind speed from them at these locations.
        """
        cli_path = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "octue", "cli.py")
        self.set_template("template-child-services")

        elevation_service_path = os.path.join(self.template_path, "elevation_service")
        elevation_service_uuid = str(uuid.uuid4())
        elevation_process = subprocess.Popen(
            [
                sys.executable,
                cli_path,
                "start",
                f"--app-dir={elevation_service_path}",
                f"--twine={os.path.join(elevation_service_path, 'twine.json')}",
                f"--config-dir={os.path.join(elevation_service_path, 'data', 'configuration')}",
                f"--service-id={elevation_service_uuid}",
                "--delete-topic-and-subscription-on-exit",
            ]
        )

        wind_speed_service_path = os.path.join(self.template_path, "wind_speed_service")
        wind_speed_service_uuid = str(uuid.uuid4())
        wind_speed_process = subprocess.Popen(
            [
                sys.executable,
                cli_path,
                "start",
                f"--app-dir={wind_speed_service_path}",
                f"--twine={os.path.join(wind_speed_service_path, 'twine.json')}",
                f"--config-dir={os.path.join(wind_speed_service_path, 'data', 'configuration')}",
                f"--service-id={wind_speed_service_uuid}",
                "--delete-topic-and-subscription-on-exit",
            ]
        )

        with ProcessesContextManager(processes=(elevation_process, wind_speed_process)):
            parent_service_path = os.path.join(self.template_path, "parent_service")

            # Dynamically alter the UUIDs defined in template children.json file to avoid conflicts when the same tests
            # run in parallel in the GitHub test runner using the actual Google Cloud PubSub instance. Apart from that,
            # the file remains the same so this test tests the template as closely as possible.
            with tempfile.TemporaryDirectory() as temporary_directory:

                with open(os.path.join(parent_service_path, "data", "configuration", "children.json")) as f:
                    template_children = json.load(f)

                template_children[0]["id"] = wind_speed_service_uuid
                template_children[1]["id"] = elevation_service_uuid

                test_children_path = os.path.join(temporary_directory, "children.json")
                with open(test_children_path, "w") as f:
                    json.dump(template_children, f)

                runner = Runner(
                    app_src=parent_service_path,
                    twine=os.path.join(parent_service_path, "twine.json"),
                    children=test_children_path,
                )
                time.sleep(5)
                analysis = runner.run(
                    input_values=os.path.join(parent_service_path, "data", "input", "values.json"),
                )

        analysis.finalise()
        self.assertTrue("elevations" in analysis.output_values)
        self.assertTrue("wind_speeds" in analysis.output_values)
